import asyncio
import hashlib
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Iterable, List, Optional

import aiofiles
import httpx

from . import logger
from .client import client_factory
from .exceptions import Ops2debFetcherError
from .parser import RemoteFile
from .utils import log_and_raise

DEFAULT_CACHE_DIRECTORY = Path("/tmp/ops2deb_cache")


async def _run(*args: str) -> asyncio.subprocess.Process:
    logger.debug(f"Running {' '.join(args)}")
    proc = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.communicate()
    return proc


async def _download_file(url: str, download_path: Path) -> None:
    tmp_path = f"{download_path}.part"
    logger.info(f"Downloading {download_path.name}...")
    try:
        async with client_factory() as client:
            async with client.stream("GET", url) as r:
                if 400 <= r.status_code < 600:
                    log_and_raise(
                        Ops2debFetcherError(
                            f"Failed to download {url}. "
                            f"Server responded with {r.status_code}."
                        )
                    )
                async with aiofiles.open(tmp_path, "wb") as f:
                    async for chunk in r.aiter_bytes():
                        await f.write(chunk)
    except httpx.HTTPError as e:
        log_and_raise(Ops2debFetcherError(f"Failed to download {url}. {str(e)}"))
    shutil.move(tmp_path, download_path)


async def _hash_file(file_path: Path) -> str:
    logger.info(f"Computing checksum for {file_path.name}...")
    sha256_hash = hashlib.sha256()
    with file_path.open("rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
            await asyncio.sleep(0)
    return sha256_hash.hexdigest()


async def _extract_archive(archive_path: Path, extract_path: Path) -> bool:
    tmp_extract_path = f"{extract_path}_tmp"
    Path(tmp_extract_path).mkdir(exist_ok=True)
    commands = [
        (
            {".tar.gz", ".tar.xz", ".tar"},
            ["/bin/tar", "-C", tmp_extract_path, "-xf", str(archive_path)],
        ),
        ({".zip"}, ["/usr/bin/unzip", "-d", tmp_extract_path, str(archive_path)]),
    ]
    selected_command: Optional[List[str]] = None

    for extensions, command in commands:
        for extension in extensions:
            if archive_path.name.endswith(extension):
                selected_command = command
                break

    if selected_command is None:
        return False

    logger.info(f"Extracting {archive_path.name}...")
    proc = await _run(*selected_command)
    if proc.returncode:
        log_and_raise(Ops2debFetcherError(f"Failed to extract archive {archive_path}"))

    shutil.move(tmp_extract_path, extract_path)
    return True


@dataclass(frozen=True)
class FetchResult:
    sha256_sum: str
    storage_path: Path


@dataclass(frozen=True)
class FetchResults:
    successes: Dict[str, FetchResult] = field(default_factory=dict)
    errors: Dict[str, Ops2debFetcherError] = field(default_factory=dict)


class FetchTask:
    def __init__(self, cache_directory_path: Path, remote_file: RemoteFile):
        url_hash = hashlib.sha256(remote_file.url.encode()).hexdigest()
        self.url = remote_file.url
        self.file_name = remote_file.url.split("/")[-1]
        self.base_path = cache_directory_path / url_hash
        self.download_path = self.base_path / self.file_name
        self.extract_path = self.base_path / f"{self.file_name}_out"
        self.checksum_path = self.base_path / f"{self.file_name}.sum"
        self.expected_hash = remote_file.sha256

    async def fetch(self, extract: bool) -> FetchResult:
        self.base_path.mkdir(exist_ok=True, parents=True)
        storage_path = self.download_path
        if self.download_path.is_file() is False:
            await _download_file(self.url, self.download_path)

        if self.checksum_path.is_file() is False:
            computed_hash = await _hash_file(self.download_path)
            self.checksum_path.write_text(computed_hash)
        else:
            computed_hash = self.checksum_path.read_text()

        if extract is True:
            if computed_hash != self.expected_hash:
                log_and_raise(
                    Ops2debFetcherError(
                        f"Wrong checksum for file {self.file_name}. "
                        f"Expected {self.expected_hash}, got {computed_hash}."
                    )
                )
            if self.extract_path.exists() is False:
                if await _extract_archive(self.download_path, self.extract_path) is True:
                    storage_path = self.extract_path
            else:
                storage_path = self.extract_path
        logger.info(f"Done with {self.download_path.name}")
        return FetchResult(computed_hash, storage_path)


class Fetcher:
    cache_directory_path = DEFAULT_CACHE_DIRECTORY

    @classmethod
    def set_cache_directory(cls, path: Path) -> None:
        cls.cache_directory_path = path

    def __init__(self, remote_files: Iterable[RemoteFile]) -> None:
        self.fetch_tasks: Dict[str, FetchTask] = {}
        for remote_file in remote_files:
            self.fetch_tasks[remote_file.url] = FetchTask(
                self.cache_directory_path, remote_file
            )

    async def fetch(self, extract: bool = True) -> FetchResults:
        logger.title(f"Fetching {len(self.fetch_tasks.keys())} files...")
        task_results = await asyncio.gather(
            *[task.fetch(extract) for task in self.fetch_tasks.values()],
            return_exceptions=True,
        )
        results = FetchResults()
        for url, task_result in zip(self.fetch_tasks.keys(), task_results):
            if isinstance(task_result, FetchResult):
                results.successes[url] = task_result
            elif isinstance(task_result, Ops2debFetcherError):
                results.errors[url] = task_result
            else:
                raise task_result
        return results

    def sync_fetch(self, extract: bool = True) -> FetchResults:
        results = asyncio.run(self.fetch(extract))
        return results



from typing import Any

import httpx


def client_factory(**kwargs: Any) -> httpx.AsyncClient:
    transport = httpx.AsyncHTTPTransport(retries=1)
    return httpx.AsyncClient(transport=transport, follow_redirects=True, **kwargs)



import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

import httpx
import ruamel.yaml
from ruamel.yaml.emitter import Emitter
from semver.version import Version

from . import logger
from .client import client_factory
from .exceptions import Ops2debUpdaterError
from .fetcher import Fetcher, FetchTask
from .parser import Blueprint, RemoteFile, load, validate
from .utils import log_and_raise, split_successes_from_errors


# fixme: move this somewhere else, this code is also duplicated in formatter.py
class FixIndentEmitter(Emitter):
    def expect_block_sequence(self) -> None:
        self.increase_indent(flow=False, indentless=False)
        self.state = self.expect_first_block_sequence_item


@dataclass(frozen=True)
class NewRelease:
    name: str
    sha256: str
    old_version: str
    new_version: str


async def _bump_and_poll(
    client: httpx.AsyncClient,
    blueprint: Blueprint,
    version: Version,
    bump_patch: bool = False,
) -> Version:
    new_version = version
    while True:
        version = version.bump_patch() if bump_patch else version.bump_minor()
        if (remote_file := blueprint.render(version=str(version)).fetch) is None:
            break
        url = remote_file.url
        logger.debug(f"Trying {url}")
        try:
            response = await client.head(url)
        except httpx.HTTPError as e:
            log_and_raise(Ops2debUpdaterError(f"Failed HEAD request to {url}. {str(e)}"))
        status = response.status_code
        if status >= 500:
            log_and_raise(Ops2debUpdaterError(f"Server error when requesting {url}"))
        if status >= 400:
            break
        else:
            new_version = version
    return new_version


async def _find_latest_release(
    blueprint: Blueprint,
) -> Optional[NewRelease]:

    if not Version.isvalid(blueprint.version):
        logger.warning(f"{blueprint.name} is not using semantic versioning")
        return None

    old_version = version = Version.parse(blueprint.version)
    async with client_factory() as client:
        version = await _bump_and_poll(client, blueprint, version, False)
        version = await _bump_and_poll(client, blueprint, version, True)

    if version != old_version:
        logger.info(f"{blueprint.name} can be bumped from {old_version} to {version}")

        file = cast(RemoteFile, blueprint.render(version=str(version)).fetch)
        result = await FetchTask(Fetcher.cache_directory_path, file).fetch(extract=False)
        return NewRelease(
            name=blueprint.name,
            sha256=result.sha256_sum,
            old_version=blueprint.version,
            new_version=str(version),
        )

    return None


async def _update_blueprint_dict(blueprint_dict: Dict[str, Any]) -> Optional[NewRelease]:
    blueprint = Blueprint.parse_obj(blueprint_dict)
    if blueprint.fetch is None:
        return None

    release = await _find_latest_release(blueprint)
    if release is None:
        return None

    blueprint_dict["version"] = release.new_version
    blueprint_dict["fetch"]["sha256"] = release.sha256
    blueprint_dict.pop("revision", None)
    return release


def update(
    configuration_path: Path, dry_run: bool = False, output_path: Optional[Path] = None
) -> None:
    yaml = ruamel.yaml.YAML(typ="rt")
    yaml.Emitter = FixIndentEmitter

    configuration_dict = load(configuration_path, yaml)
    validate(configuration_dict)

    logger.title("Looking for new releases...")

    # configuration file can be a list of blueprints or a single blueprint
    blueprints_dict = (
        configuration_dict
        if isinstance(configuration_dict, list)
        else [configuration_dict]
    )

    async def run_tasks() -> Any:
        return await asyncio.gather(
            *[_update_blueprint_dict(b) for b in blueprints_dict],
            return_exceptions=True,
        )

    results: List[Union[NewRelease, Exception]] = asyncio.run(run_tasks())
    new_releases, errors = split_successes_from_errors(results)

    if dry_run is False and new_releases:
        with configuration_path.open("w") as output:
            yaml.dump(configuration_dict, output)
        logger.info("Configuration file updated")

        if output_path is not None:
            lines = [
                f"Updated {r.name} from {r.old_version} to {r.new_version}"
                for r in new_releases
            ]
            output_path.write_text("\n".join(lines) + "\n")

    if not new_releases:
        logger.info("Did not found any updates")

    if errors:
        raise Ops2debUpdaterError(f"{len(errors)} failures occurred")



import asyncio
from dataclasses import dataclass
from itertools import chain, product
from typing import Iterator, List

import httpx
from debian.deb822 import Packages, Release
from pydantic import BaseModel, Field, HttpUrl, ValidationError

from . import logger
from .client import client_factory
from .exceptions import Ops2debAptError


class DebianRepository(BaseModel):
    url: HttpUrl
    distribution: str = Field(..., regex=r"[a-zA-Z0-9]+")


@dataclass
class DebianRepositoryPackage:
    name: str
    version: str
    architecture: str


async def _download_repository_release_file(
    client: httpx.AsyncClient, distribution: str
) -> Release:
    url = f"/dists/{distribution}/Release"
    response = await client.get(url)
    response.raise_for_status()
    return Release(response.text)


async def _download_repository_packages_file(
    client: httpx.AsyncClient, distribution: str, component: str, arch: str
) -> bytes:
    url = f"/dists/{distribution}/{component}/binary-{arch}/Packages"
    logger.debug(f"Downloading {url}...")
    async with client.stream("GET", url) as response:
        response.raise_for_status()
        return await response.aread()


def _parse_repository_packages_file(content: bytes) -> Iterator[DebianRepositoryPackage]:
    """Extract package names and versions from a repo Packages file"""
    for src in Packages.iter_paragraphs(content, use_apt_pkg=False):
        yield DebianRepositoryPackage(src["Package"], src["Version"], src["Architecture"])


def _parse_debian_repository_option(debian_repository: str) -> DebianRepository:
    try:
        url, distribution = debian_repository.split(" ")
    except ValueError:
        raise Ops2debAptError(
            "The expected format for the --repository option is "
            '"{repository_url} {distribution}"'
        )
    try:
        return DebianRepository(url=url, distribution=distribution)
    except ValidationError as e:
        raise Ops2debAptError(str(e))


async def list_repository_packages(
    debian_repository: str,
) -> List[DebianRepositoryPackage]:
    repository = _parse_debian_repository_option(debian_repository)
    async with client_factory(base_url=repository.url) as client:
        release = await _download_repository_release_file(client, repository.distribution)
        architectures = release["Architectures"].split(" ")
        components = release["Components"].split(" ")
        logger.info(
            f"Repository {repository.url} {repository.distribution} has architectures "
            f"{architectures} and components {components}"
        )
        tasks = [
            _download_repository_packages_file(
                client, repository.distribution, component, architecture
            )
            for component, architecture in product(components, architectures)
        ]
        results = [
            _parse_repository_packages_file(r) for r in await asyncio.gather(*tasks)
        ]
    return list(chain(*results))


def sync_list_repository_packages(
    debian_repository: str,
) -> List[DebianRepositoryPackage]:
    """Example: "http://deb.wakemeops.com/ stable" """
    return asyncio.run(list_repository_packages(debian_repository))



from typer import colors, secho

_debug: bool = False


def enable_debug(enable: bool) -> None:
    global _debug
    _debug = enable


def info(message: str) -> None:
    secho(message)


def debug(message: str) -> None:
    if _debug is True:
        secho(message, fg=colors.BRIGHT_BLACK)


def warning(message: str) -> None:
    secho(message, fg=colors.YELLOW)


def error(message: str) -> None:
    secho(message, fg=colors.RED)


def title(message: str) -> None:
    secho(message, fg=colors.BLUE, bold=True)



import shutil
import subprocess
from pathlib import Path
from typing import List, Optional

from jinja2 import Environment, FunctionLoader

from . import logger
from .apt import DebianRepositoryPackage, sync_list_repository_packages
from .exceptions import Ops2debGeneratorError, Ops2debGeneratorScriptError
from .fetcher import Fetcher, FetchResult
from .parser import Blueprint
from .templates import template_loader

_environment = Environment(loader=FunctionLoader(template_loader))


def _format_command_output(output: str) -> str:
    lines = output.splitlines()
    output = "\n  ".join([line for line in lines])
    return "> " + output


class SourcePackage:
    def __init__(self, blueprint: Blueprint, output_directory: Path):
        self.directory_name = f"{blueprint.name}_{blueprint.version}_{blueprint.arch}"
        self.package_directory = (output_directory / self.directory_name).absolute()
        self.debian_directory = self.package_directory / "debian"
        self.src_directory = self.package_directory / "src"
        self.fetch_directory = self.package_directory / "fetched"
        self.blueprint = blueprint.render(self.src_directory)
        self.debian_version = (
            f"{self.blueprint.version}-{self.blueprint.revision}~ops2deb"
        )

    def _render_tpl(self, template_name: str) -> None:
        template = _environment.get_template(f"{template_name}")
        package = self.blueprint.dict(exclude={"fetch", "script"})
        package.update({"version": self.debian_version})
        template.stream(package=package).dump(str(self.debian_directory / template_name))

    def _init(self) -> None:
        """Reset source package directory"""
        # without artifacts from previous builds
        shutil.rmtree(self.debian_directory, ignore_errors=True)
        self.debian_directory.mkdir(parents=True)
        shutil.rmtree(self.fetch_directory, ignore_errors=True)
        shutil.rmtree(self.src_directory, ignore_errors=True)
        self.src_directory.mkdir(parents=True)
        for path in ["usr/bin", "usr/share", "usr/lib"]:
            (self.src_directory / path).mkdir(parents=True)

    def _populate_with_fetch_result(self, fetch_result: FetchResult) -> None:

        if fetch_result.storage_path.is_file():
            self.fetch_directory.mkdir(exist_ok=True)
            shutil.copy2(fetch_result.storage_path, self.fetch_directory)
        else:
            shutil.copytree(fetch_result.storage_path, self.fetch_directory)

    def _run_script(self) -> None:
        # if blueprint has no fetch instruction, we stay in the directory from which
        # ops2deb was called
        cwd = self.fetch_directory if self.blueprint.fetch else Path(".")

        # run script
        for line in self.blueprint.script:
            logger.info(f"$ {line}")
            result = subprocess.run(line, shell=True, cwd=cwd, capture_output=True)
            if stdout := result.stdout.decode():
                logger.info(_format_command_output(stdout))
            if stderr := result.stderr.decode():
                logger.error(_format_command_output(stderr))
            if result.returncode:
                raise Ops2debGeneratorScriptError

    def generate(self, fetch_result: Optional[FetchResult] = None) -> None:
        logger.title(f"Generating source package {self.directory_name}...")

        # fetch failed, we cannot generate source package
        if self.blueprint.fetch is not None and fetch_result is None:
            return

        # make sure we generate source packages in a clean environment
        # without artifacts from previous builds
        self._init()

        # copy downloaded/extracted archive to package fetch directory
        if fetch_result is not None:
            self._populate_with_fetch_result(fetch_result)

        # render debian/* files
        for template in [
            "changelog",
            "control",
            "rules",
            "compat",
            "install",
            "lintian-overrides",
        ]:
            self._render_tpl(template)

        # run blueprint script
        self._run_script()


def filter_already_published_packages(
    packages: List[SourcePackage], debian_repository: str
) -> List[SourcePackage]:
    already_published_packages = sync_list_repository_packages(debian_repository)
    filtered_packages: List[SourcePackage] = []
    for package in packages:
        if (
            DebianRepositoryPackage(
                package.blueprint.name, package.debian_version, package.blueprint.arch
            )
            not in already_published_packages
        ):
            filtered_packages.append(package)
    return filtered_packages


def generate(
    blueprints: List[Blueprint],
    output_directory: Path,
    debian_repository: Optional[str] = None,
) -> None:

    packages = [SourcePackage(blueprint, output_directory) for blueprint in blueprints]

    # filter out blueprints that build packages already available in the debian repository
    if debian_repository is not None:
        packages = filter_already_published_packages(packages, debian_repository)

    # run fetch instructions (download, verify, extract) in parallel
    files = [p.blueprint.fetch for p in packages if p.blueprint.fetch is not None]
    results = Fetcher(files).sync_fetch()

    for package in packages:
        if package.blueprint.fetch is not None:
            package.generate(results.successes.get(package.blueprint.fetch.url))
        else:
            package.generate()

    if results.errors:
        raise Ops2debGeneratorError(f"{len(results.errors)} failures occurred")



from typing import Optional

DEBIAN_CHANGELOG = """\
{{ package.name }} ({{ package.version }}) stable; urgency=medium

  * Release {{ package.version }}

 -- ops2deb <ops2deb@upciti.com>  Tue, 07 May 2019 20:31:30 +0000
"""

DEBIAN_COMPAT = """\
10
"""

DEBIAN_CONTROL = """\
Source: {{ package.name }}
Priority: optional
Maintainer: ops2deb <ops2deb@upciti.com>
Build-Depends: debhelper
Standards-Version: 3.9.6
{%- if package.homepage %}{{ '\n' }}Homepage: {{ package.homepage }}{% endif %}

Package: {{ package.name }}
Architecture: {{ package.arch }}
{%- if package.depends %}{{ '\n' }}Depends: {{ package.depends|sort|join(', ') }}{% endif %}
{%- if package.recommends %}{{ '\n' }}Recommends: {{ package.recommends|sort|join(', ') }}{% endif %}
{%- if package.conflicts %}{{ '\n' }}Conflicts: {{ package.conflicts|sort|join(', ') }}{% endif %}
Description: {{ package.summary }}
 {{ package.description | indent(1) }}
"""

DEBIAN_INSTALL = """\
src/* /
"""

DEBIAN_LINTIAN_OVERRIDES = """\
{{ package.name }}: statically-linked-binary
{{ package.name }}: binary-without-manpage
"""

DEBIAN_RULES = """\
#!/usr/bin/make -f

%:
	dh $@

override_dh_shlibdeps:
	true

override_dh_strip:
	dh_strip --no-ddebs

override_dh_builddeb:
	dh_builddeb -- -Zxz
"""


def template_loader(name: str) -> Optional[str]:
    variable_name = f"DEBIAN_{name.upper().replace('-', '_')}"
    template_content: str = globals()[variable_name]
    if variable_name in globals():
        return template_content
    return None



from typing import Iterable, List, Tuple, TypeVar, Union

from . import logger
from .exceptions import Ops2debError


def log_and_raise(exception: Exception) -> None:
    logger.error(str(exception))
    raise exception


T = TypeVar("T")


def split_successes_from_errors(
    items: Iterable[Union[T, Exception]]
) -> Tuple[List[T], List[Ops2debError]]:
    results: List[T] = []
    errors: List[Ops2debError] = []
    for item in items:
        if isinstance(item, Ops2debError):
            errors.append(item)
        elif isinstance(item, Exception):
            raise item
        else:
            results.append(item)
    return results, errors



import asyncio
import re
from pathlib import Path
from typing import Any, Dict, Optional

from . import logger
from .exceptions import Ops2debBuilderError
from .utils import log_and_raise


def parse_debian_control(cwd: Path) -> Dict[str, str]:
    """
    Extract fields from debian/control
    :param cwd: Path to debian source package
    :return: Dict object with fields as keys
    """
    field_re = re.compile(r"^([\w-]+)\s*:\s*(.+)")

    content = (cwd / "debian" / "control").read_text()
    control = {}
    for line in content.split("\n"):
        m = field_re.search(line)
        if m:
            g = m.groups()
            control[g[0]] = g[1]

    return control


async def build_package(cwd: Path) -> None:
    """Run dpkg-buildpackage in specified path."""
    args = ["-us", "-uc"]
    arch = parse_debian_control(cwd)["Architecture"]
    if arch != "all":
        args += ["--host-arch", arch]

    logger.info(f"Building {cwd}...")

    proc = await asyncio.create_subprocess_exec(
        "/usr/bin/dpkg-buildpackage",
        *args,
        cwd=cwd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()

    if proc.returncode:
        log_and_raise(Ops2debBuilderError(f"Failed to build package in {str(cwd)}"))
    else:
        logger.info(f"Successfully built {str(cwd)}")
    if stdout:
        logger.debug(stdout.decode())
    if stderr:
        logger.debug(stderr.decode())


def build(output_directory: Path, workers: int = 4) -> None:
    """
    Run several instances of dpkg-buildpackage in parallel.
    :param output_directory: path where to search for source packages
    :param workers: Number of threads to run in parallel
    """

    logger.title("Building source packages...")

    paths = []
    for output_directory in output_directory.iterdir():
        if output_directory.is_dir() and (output_directory / "debian/control").is_file():
            paths.append(output_directory)

    async def _build_package(sem: asyncio.Semaphore, _path: Path) -> Optional[int]:
        async with sem:  # semaphore limits num of simultaneous builds
            return await build_package(_path)

    async def _build_all() -> Any:
        sem = asyncio.Semaphore(workers)
        return await asyncio.gather(
            *[_build_package(sem, p) for p in paths], return_exceptions=True
        )

    results = asyncio.run(_build_all())

    if errors := [e for e in results if isinstance(e, Exception)]:
        raise Ops2debBuilderError(f"{len(errors)} failures occurred")



class Ops2debError(Exception):
    pass


class Ops2debGeneratorError(Ops2debError):
    pass


class Ops2debGeneratorScriptError(Ops2debGeneratorError):
    pass


class Ops2debBuilderError(Ops2debGeneratorError):
    pass


class Ops2debFetcherError(Ops2debError):
    pass


class Ops2debUpdaterError(Ops2debError):
    pass


class Ops2debParserError(Ops2debError):
    pass


class Ops2debAptError(Ops2debError):
    pass


class Ops2debFormatterError(Ops2debError):
    pass



import shutil
import sys
import traceback
from pathlib import Path
from typing import NoReturn, Optional

import typer

from . import builder, formatter, generator, logger, parser, updater
from .exceptions import Ops2debError
from .fetcher import DEFAULT_CACHE_DIRECTORY, Fetcher

# Options below are used by multiple subcommands
OPTION_CONFIGURATION = typer.Option(
    "ops2deb.yml",
    "--config",
    "-c",
    envvar="OPS2DEB_CONFIG",
    help="Path to configuration file.",
)
OPTION_CACHE_DIRECTORY: Path = typer.Option(
    DEFAULT_CACHE_DIRECTORY,
    "--cache-dir",
    envvar="OPS2DEB_CACHE_DIR",
    help="Directory where files specified in fetch instructions are downloaded.",
)
OPTION_OUTPUT_DIRECTORY: Path = typer.Option(
    "output",
    "--output-dir",
    "-o",
    envvar="OPS2DEB_OUTPUT_DIR",
    help="Directory where debian source packages are generated and built.",
)

DEFAULT_EXIT_CODE = 1
_exit_code = DEFAULT_EXIT_CODE

app = typer.Typer()


def error(exception: Exception) -> NoReturn:
    logger.error(str(exception))
    logger.debug(traceback.format_exc())
    sys.exit(_exit_code)


@app.command(help="Generate debian source packages from configuration file.")
def generate(
    configuration_path: Path = OPTION_CONFIGURATION,
    output_directory: Path = OPTION_OUTPUT_DIRECTORY,
    cache_directory: Path = OPTION_CACHE_DIRECTORY,
    debian_repository: str = typer.Option(
        None,
        "--repository",
        "-r",
        envvar="OPS2DEB_REPOSITORY",
        help='Format: "{debian_repo_url} {distribution_name}". '
        'Example: "http://deb.wakemeops.com/ stable". '
        "Packages already published in the repo won't be generated.",
    ),
) -> None:
    Fetcher.set_cache_directory(cache_directory)
    try:
        generator.generate(
            parser.parse(configuration_path), output_directory, debian_repository
        )
    except Ops2debError as e:
        error(e)


@app.command(help="Build debian packages from source packages.")
def build(output_directory: Path = OPTION_OUTPUT_DIRECTORY) -> None:
    try:
        builder.build(output_directory)
    except Ops2debError as e:
        error(e)


@app.command(help="Clear ops2deb download cache.")
def purge(cache_directory: Path = OPTION_CACHE_DIRECTORY) -> None:
    shutil.rmtree(cache_directory, ignore_errors=True)


@app.command(help="Look for new application releases and edit configuration file.")
def update(
    config: Path = OPTION_CONFIGURATION,
    cache_directory: Path = OPTION_CACHE_DIRECTORY,
    dry_run: bool = typer.Option(
        False, "--dry-run", "-d", help="Don't edit config file."
    ),
    output_path: Optional[Path] = typer.Option(
        None,
        "--output-file",
        envvar="OPS2DEB_OUTPUT_FILE",
        help="Path to file where to save a summary of updated files.",
    ),
) -> None:
    Fetcher.set_cache_directory(cache_directory)
    try:
        updater.update(config, dry_run, output_path)
    except Ops2debError as e:
        error(e)


@app.command(help="Validate configuration file.")
def validate(configuration_path: Path = OPTION_CONFIGURATION) -> None:
    try:
        parser.parse(configuration_path)
    except Ops2debError as e:
        error(e)


@app.command(help="Format configuration file.")
def format(configuration_path: Path = OPTION_CONFIGURATION) -> None:
    try:
        formatter.format(configuration_path)
    except Ops2debError as e:
        error(e)


@app.callback()
def args_cb(
    verbose: bool = typer.Option(
        False, "--verbose", "-v", envvar="OPS2DEB_VERBOSE", help="Enable more logs."
    ),
    exit_code: int = typer.Option(
        DEFAULT_EXIT_CODE,
        "--exit-code",
        "-e",
        envvar="OPS2DEB_EXIT_CODE",
        help="Exit code to use in case of failure.",
    ),
) -> None:
    global _exit_code
    _exit_code = exit_code
    if exit_code > 255 or exit_code < 0:
        raise typer.BadParameter("Invalid exit code")
    logger.enable_debug(verbose)


def main() -> None:
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ops2deb']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3,<4',
 'PyYAML>=6,<7',
 'aiofiles>=0.1.9',
 'httpx>=0.20.0',
 'pydantic>=1,<2',
 'python-debian>=0.1.42',
 'ruamel.yaml>=0.17.16',
 'semver==3.0.0-dev.2',
 'typer>=0.4.0']

entry_points = \
{'console_scripts': ['ops2deb = ops2deb.cli:main']}

setup_kwargs = {
    'name': 'ops2deb',
    'version': '0.15.0',
    'description': 'Build debian packages',
    'long_description': "![cicd](https://github.com/upciti/ops2deb/actions/workflows/cicd.yml/badge.svg)\n[![codecov](https://codecov.io/gh/upciti/ops2deb/branch/main/graph/badge.svg)](https://codecov.io/gh/upciti/ops2deb)\n[![MIT license](https://img.shields.io/badge/License-MIT-blue.svg)](https://lbesson.mit-license.org/)\n[![Generic badge](https://img.shields.io/badge/type_checked-mypy-informational.svg)](https://mypy.readthedocs.io/en/stable/introduction.html)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![PyPI version shields.io](https://img.shields.io/pypi/v/ops2deb.svg)](https://pypi.python.org/pypi/ops2deb/)\n[![Downloads](https://static.pepy.tech/personalized-badge/ops2deb?period=total&units=international_system&left_color=blue&right_color=green&left_text=Downloads)](https://pepy.tech/project/ops2deb)\n\n# ops2deb\n\nAre you tired of checking if your favorite devops tools are up-to-date? Are you using a debian based GNU/Linux distribution?\n`ops2deb` is designed to generate Debian packages for common devops tools such as kubectl, kustomize, helm, ...,\nbut it could be used to package any statically linked application. In short, it consumes a configuration file and outputs `.deb` packages.\n\n## Configuration file\n\nWritten in YAML and composed of a single blueprint object or a list of blueprints objects. A blueprint is defined by the following:\n\n| Field         | Meaning                                                                                                                                     | Default |\n| ------------- | ------------------------------------------------------------------------------------------------------------------------------------------- | ------- |\n| `name`        | Component name, e.g. `kustomize`.                                                                                                           |         |\n| `version`     | Application release to package.                                                                                                             |         |\n| `homepage`    | Upstream project homepage.                                                                                                                  | `None`  |\n| `arch`        | Package architecture.                                                                                                                       | `amd64` |\n| `revision`    | Package revistion.                                                                                                                          | `1`     |\n| `summary`     | Package short description.                                                                                                                  |         |\n| `description` | Package full description.                                                                                                                   |         |\n| `fetch`       | A binary to download, and a `sha256` checksum. `tar.gz`, `tar.xz`, `tar` and `zip` (requires `unzip`) archives are extracted automatically. | `Null`  |\n| `script`      | List of build instructions templated with jinja2 and intepreted with the default `shell`.                                                   | `[]`    |\n| `depends`     | List of package dependencies. Corresponds to `Depends` entry in `debian/control`.                                                           | `[]`    |\n| `recommends`  | List of package recommended dependencies. Corresponds to `Recommends` entry in `debian/control`.                                            | `[]`    |\n| `conflicts`   | List of conflicting packages. Corresponds to `Conflicts` entry in `debian/control`.                                                         | `[]`    |\n\nExample of a configuration file a single blueprint:\n\n```yaml\nname: kubectl\nversion: 1.20.1\nsummary: Command line client for controlling a Kubernetes cluster\ndescription: |\n  kubectl is a command line client for running commands against Kubernetes clusters.\nfetch:\n  url: https://storage.googleapis.com/kubernetes-release/release/v{{version}}/bin/linux/amd64/kubectl\n  sha256: 3f4b52a8072013e4cd34c9ea07e3c0c4e0350b227e00507fb1ae44a9adbf6785\nscript:\n  - mv kubectl {{src}}/usr/bin/\n```\n\n## Dependencies\n\n- Python >= 3.9\n- To build debian packages with `ops2deb build` you need the following packages on your host:\n\n```shell\nsudo apt install build-essential fakeroot debhelper\n```\n\n## Installation\n\nInstall `ops2deb` in a virtualenv or with [pipx](https://github.com/pipxproject/pipx)\n\n```shell\npipx install ops2deb\n```\n\n## Getting started\n\nIn a test directory run:\n\n```shell\ncurl https://raw.githubusercontent.com/upciti/ops2deb/main/ops2deb.yml\nops2deb generate\nops2deb build\n```\n\nTo check for new releases run:\n\n```shell\nops2deb update\n```\n\nThis command updates each blueprint in the `ops2deb.yml` configuration file with the latest version of\nthe upstream application (currently only works for applications using semantic versioning).\n\nBy default `ops2deb` caches downloaded content in `/tmp/ops2deb_cache`:\n\n```shell\ntree /tmp/ops2deb_cache\n```\n\nThe cache can be flushed with:\n\n```shell\nops2deb purge\n```\n\nFor more information about existing subcommands and options run `ops2deb --help`.\n\n## Usage examples\n\n### Creating a metapackage\n\nOps2deb can be used to create [metapackages](https://www.debian.org/blends/hamradio/get/metapackages):\n\n```yaml\nname: allthethings\nversion: 0.1.9\nsummary: Install various devops tools\ndescription: Some great description.\ndepends:\n  - kubectl\n  - kustomize\n  - helm\n  - helmfile\n  - devspace\n```\n\n### Packaging ops2deb with ops2deb\n\nNote that when the fetch key is not used, ops2deb will run the build script from the directory where it was called.\nHence for the following blueprint to succeed, you have to run ops2deb from the root directory of this github project.\n\n```yaml\nname: ops2deb\nversion: 0.15.0\nhomepage: https://github.com/upciti/ops2deb\nsummary: Debian packaging tool for portable applications\ndescription: |-\n  Ops2deb is primarily designed to easily generate Debian packages for portable\n  applications such as single binary applications and scripts. Packages are\n  described using a simple configuration file format. Ops2deb can track new\n  releases of upstream applications and automatically bump application versions\n  in its configuration file.\nscript:\n  - ./build-single-binary-application.sh\n  - install -m 755 build/x86_64-unknown-linux-gnu/release/install/ops2deb {{src}}/usr/bin/\n```\n\n## Development\n\nYou will need [poetry](https://python-poetry.org/), and probably [pyenv](https://github.com/pyenv/pyenv) if you don't have python 3.9 on your host.\n\n```shell\npoetry install\n```\n\nTo run ops2deb test suite run:\n\n```shell\npoetry run task check\n```\n\n## Important notes\n\n`ops2deb` **DOES NOT** sandbox build instructions so if you do something like:\n\n```shell\nscript:\n- rm -rf ~/*\n```\n\nYou will loose your files... To make sure that you won't mess with your system, run it within a container.\n",
    'author': 'Upciti',
    'author_email': 'support@upciti.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/upciti/ops2deb',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import json
from operator import attrgetter
from pathlib import Path
from textwrap import wrap
from typing import Any, Dict, List

import yaml

from .exceptions import Ops2debFormatterError
from .parser import Blueprint, parse


class PrettyYAMLDumper(yaml.dumper.SafeDumper):
    def expect_block_sequence(self) -> None:
        self.increase_indent(flow=False, indentless=False)
        self.state = self.expect_first_block_sequence_item

    def choose_scalar_style(self) -> str:
        style = super().choose_scalar_style()
        style = '"' if style == "'" else style
        style = "|" if self.analysis.multiline else style
        return style


def sort_blueprints(blueprints: List[Blueprint]) -> List[Blueprint]:
    return sorted(blueprints, key=attrgetter("name", "version", "revision"))


def format_description(description: str) -> str:
    lines: List[str] = []
    for line in description.split("\n"):
        lines.extend(wrap(line, width=79))
    return "\n".join(lines)


def format_blueprint(blueprint: Blueprint) -> Dict[str, Any]:
    update: Dict[str, str] = {"description": format_description(blueprint.description)}
    blueprint = blueprint.copy(update=update)
    keys_with_list = {"recommends", "conflicts", "depends", "script"}
    exclude = {k for k in keys_with_list if not getattr(blueprint, k)}
    return json.loads(blueprint.json(exclude_defaults=True, exclude=exclude))


def format(configuration_path: Path) -> None:
    original_configuration_content = configuration_path.read_bytes()

    # sort blueprints by name, version and revision
    blueprints = sort_blueprints(parse(configuration_path))

    # wrap descriptions, remove default values, remove empty lists
    formatted_blueprints = [format_blueprint(b) for b in blueprints]

    # dump to yaml, use | for multiline strings and double quotes instead of single quotes
    yaml_dump = yaml.dump(
        formatted_blueprints if len(blueprints) > 1 else formatted_blueprints[0],
        Dumper=PrettyYAMLDumper,
        default_flow_style=False,
        sort_keys=False,
        encoding="utf-8",
    )

    # add line break between blueprints
    yaml_dump_lines = yaml_dump.split(b"\n")
    new_yaml_dump_lines: List[bytes] = [yaml_dump_lines[0]]
    for line in yaml_dump_lines[1:]:
        if line.startswith(b"- "):
            new_yaml_dump_lines.append(b"")
        new_yaml_dump_lines.append(line)

    # save formatted configuration file
    formatted_configuration_content = b"\n".join(new_yaml_dump_lines)
    configuration_path.write_bytes(formatted_configuration_content)

    if formatted_configuration_content != original_configuration_content:
        raise Ops2debFormatterError(f"Reformatted {configuration_path}")



from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

from jinja2 import Environment
from pydantic import AnyHttpUrl, BaseModel, Field, ValidationError
from ruamel.yaml import YAML, YAMLError

from .exceptions import Ops2debParserError

environment = Environment()
Architecture = Literal["all", "amd64", "armhf"]


class Base(BaseModel):
    class Config:
        extra = "forbid"
        allow_mutation = False


class RemoteFile(Base):
    url: AnyHttpUrl = Field(..., description="File URL")
    sha256: str = Field(..., description="File SHA256 checksum")


class Blueprint(Base):
    name: str = Field(..., description="Package name")
    version: str = Field(..., description="Package name")
    homepage: Optional[AnyHttpUrl] = Field(None, description="Upstream project homepage")
    revision: int = Field(1, description="Package revision")
    arch: Architecture = Field("amd64", description="Package architecture")
    summary: str = Field(..., description="Package short description, one line only")
    description: str = Field(..., description="Package description")
    depends: List[str] = Field(default_factory=list, description="Package dependencies")
    recommends: List[str] = Field(
        default_factory=list, description="Package recommended dependencies"
    )
    conflicts: List[str] = Field(
        default_factory=list,
        description="Conflicting packages, for more information read "
        "https://www.debian.org/doc/debian-policy/ch-relationships.html",
    )
    fetch: Optional[RemoteFile] = Field(None, description="File to download")
    script: List[str] = Field(default_factory=list, description="Build instructions")

    class Config:
        anystr_strip_whitespace = True

    def _render_str(self, string: str, **kwargs: Optional[str]) -> str:
        version = kwargs.pop("version", None)
        version = version or self.version
        return environment.from_string(string).render(
            name=self.name,
            arch=self.arch,
            version=version,
            **kwargs,
        )

    def render(
        self, src: Optional[Path] = None, version: Optional[str] = None
    ) -> "Blueprint":
        update: Dict[str, Any] = {}

        if src is not None:
            update["script"] = [
                self._render_str(line, src=str(src)) for line in self.script
            ]

        if self.fetch is not None:
            update["fetch"] = RemoteFile(
                url=self._render_str(self.fetch.url, version=version),
                sha256=self.fetch.sha256,
            )

        return self.copy(update=update)


class Configuration(Base):
    __root__: Union[List[Blueprint], Blueprint]


def load(
    configuration_path: Path, yaml: YAML = YAML()
) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    try:
        return yaml.load(configuration_path.open("r"))
    except YAMLError as e:
        raise Ops2debParserError(f"Invalid YAML file.\n{e}")
    except FileNotFoundError:
        raise Ops2debParserError(f"File not found: {configuration_path.absolute()}")


def validate(
    configuration_dict: Union[List[Dict[str, Any]], Dict[str, Any]]
) -> List[Blueprint]:
    try:
        blueprints = Configuration.parse_obj(configuration_dict).__root__
    except ValidationError as e:
        raise Ops2debParserError(f"Invalid configuration file.\n{e}")
    if isinstance(blueprints, Blueprint):
        blueprints = [blueprints]
    return blueprints


def parse(configuration_path: Path) -> List[Blueprint]:
    return validate(load(configuration_path))
